from OFS import SimpleItem
from Globals import HTMLFile
import ConflictResolvingTools

def manage_addZQueueMessage(self, id, message, fromPacket=1):
    '''Add a ZQueueTopic to the self folder.'''
    self._setObject(id, ZQueueMessage(id, message, fromPacket))


class ZQueueMessage(SimpleItem.SimpleItem):
    '''ZQueue Message Object'''
    meta_type = "ZQueueMessage"
    
    def __init__(self, id, message, packaged=1):
        '''Initialiser'''
        self.id = id
        self.message = message
        self.wrappedInPackage = packaged
        self.subscribersReceived = {}
        self.numberOfSubscribersReceived = 0

    def _p_resolveConflict(self, oldState, savedState, newState):
        '''Conflict resolver.'''
        try:
            #set up a handler with a default handler that raises an error if the values are different
            handler = ConflictResolvingTools.KeyConflictHandler(ConflictResolvingTools.errorIfDifferent)
            handler.addHandleableKey('_objects', ConflictResolvingTools.tupleMerge)
            handler.addHandleableKey('numberOfSubscribersReceived', ConflictResolvingTools.returnLargest)
            handler.addHandleableKey('subscribersReceived', ConflictResolvingTools.dictMerge)
            handler.addHandleableKey('wrappedInPackage', ConflictResolvingTools.returnLargest)
            retState = ConflictResolvingTools.mergeDiffs(oldState, savedState, newState, handler)
            #manually calculate the numberOfSubscribersReceived
            retState['numberOfSubscribersReceived'] = len(retState['subscribersReceived'].keys())
            #make sure that there are keys for every id in _objects.  Remove those that are not relevant
            ConflictResolvingTools.synchroniseObjects(retState)
            return retState
        except:
            tb = Helpers.GetTraceback()
            print "!" * 120
            print "Error resolving conflict for Message"
            print tb
            return None
        
    def addSubsriberReceived(self, subscriberId):
        '''Add to the list of subscribers that have received this message.'''
        self.subscribersReceived[subscriberId] = 1
        self.numberOfSubscribersReceived += 1
        self._p_changed = 1
        
    def isPackaged(self):
        '''Return if this message have been put in a packet for delivery.'''
        return self.wrappedInPackage
    
    def setIsPackaged(self):
        '''Sets this packet as having been put in a packet for delivery.'''
        self.wrappedInPackage = 1
        self._p_changed
        
    def subscriberReceived(self, subsriberId):
        '''Return true if the subscriberId has already received the message.'''
        return self.subscribersReceived.has_key(subsriberId)
    
    def getNumberOfSubscribersReceived(self):
        '''Returns the number of subscribers that have received this message.'''
        return self.numberOfSubscribersReceived
        
    def setMessage(self, message):
        '''Sets the message.'''
        self.message = message
        self._p_changed = 1
        
    def getMessage(self):
        '''Returns the message.'''
        return self.message
        
    def index_html(self):
        '''Admin interface for message.'''
        messageId = self.id
        messageBody = self.message
        subscribersReceivedList = []
        for subscriberReceived in self.subscribersReceived.keys():
            messageList.append('''%s<br>''' %(subscriberReceived))
            
        subscribersReceivedList = "\n".join(subscribersReceivedList)
        
        return HTMLFile('dtml/message', globals())(client=self, mapping=locals())        