from OFS import Folder, SimpleItem
import Topic, Packet
import threading, thread, Queue, pickle, StringIO, time, xmlrpclib, sys
from Helpers import Helpers
from Globals import HTMLFile
import Globals
import ConflictResolvingTools

Globals.globalLocks = {}

def getLock(id, wait=1):
    '''Get the global queue lock.
    id : A id for the lock required (a new lock will be created if nescessary)
    wait : Block if the lock is busy, otherwise return 0 if the lock is busy
    retries : times to retry to get the lock
    waitBeforeRetry : time to wait before retrying again.'''
    if not Globals.globalLocks.has_key(id):
        Globals.globalLocks[id] = threading.Lock()
    
    return Globals.globalLocks[id].acquire(wait)

def releaseLock(id):
    '''Release the global queue lock.
    id : A id for the lock required'''
    Globals.globalLocks[id].release()

def manage_addZQueueNode(self, id='ZQueueNode', nodeId='0'):
    '''Factory method to create a ZQeueNode
    id : Zope Id (default = ZQueueNode)
    nodeId : This node's system unique id (default = 0)'''
    self._setObject(id, ZQueueNode(id, nodeId))

    #XXX Do a proper redirect and create a proper manage_addForm

    return 1

class ZQueueNode(Folder.Folder):
    '''ZQueue Node object.
    A messaging system consists of ZQueueNode objects (which may be in multiple Zopes)'''

    meta_type = "ZQueueNode"

    def __init__(self, id, nodeId):
        '''ZQueueNode initialiser
        id : Zope Id
        nodeId : This node's system unique id'''
        self.id = id
        self.nodeId = nodeId
        self.remoteNodes = {}
        self.packetIdsReceived = {}
        self.currentPacketNumber = 0
        self.deletedRootPackets = []
        self.currentPacketIds = {}
        self.deliveryHandlerStarted = 0

    def clearNode(self, REQUEST=None):
        '''Reset the node contents.
        REQUEST : Zope request object (used for redirect)
        This will clear the node completely and re-create the ZQueuePoller object.
        Mainly used for testing purposes.'''
        self.manage_delObjects(self.objectIds())
        self.packetIdsReceived = {}
        self.currentPacketNumber = 0
        self.deletedRootPackets = []
        self.currentPacketIds = {}
        self._p_changed = 1
        if REQUEST:
            self.redirect(REQUEST)

    #public methods for client programs
    
    
    
    def publishMessageToTopic(self, topic, message):
        '''Publishes a message to a topic.
        topic : case sensitive name of the topic
        message : picklable, XML-RPC marshable message body
        Creates the topic if needed.'''
        try:
            published = 0
            while not published:
                try:
                    getLock(self.absolute_url())
                    try:
                        topicOb = self.getTopic(topic)
                        topicOb.publishMessage(message)
                        published = 1
                    except ConflictResolvingTools.ConflictError:
                        get_transaction().abort()
                        published = 0
                finally:
                    releaseLock(self.absolute_url())
                if not published:
                    print "Could not publish"
                    time.sleep(0.5)
        except:
            tb = Helpers.GetTraceback()
            print "!"*120
            print "publishMessageToTopic Error"
            print "topic = %s" %topic
            print "message = %s" %message
            print tb
            raise

    def subscribeToTopic(self, topic, applicationId,type=0):
        '''Subscribes an application to a topic.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        type : for type==0, a polling client (default = 0)
        Creates the topic if needed.'''
        subscribed = 0
        while not subscribed:
            try:
                getLock(self.absolute_url())
                try:
                    topicOb = self.getTopic(topic)
                    topicOb.subscribe(applicationId,type)
                    subscribed = 1
                except ConflictResolvingTools.ConflictError:
                    get_transaction().abort()
                    subscribed = 0
            finally:
                releaseLock(self.absolute_url())
            if not subscribed:
                print "Could not subscribe"
                time.sleep(0.5)

    def unsubscribeFromTopic(self, topic, applicationId):
        '''Unsubscribe an application from topic.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        '''
        unsubscribed = 0
        while not unsubscribed:
            try:
                getLock(self.absolute_url())
                try:
                    topicOb = getattr(self, topic)
                    topicOb.unsubscribe(applicationId)
                    unsubscribed = 1
                except ConflictResolvingTools.ConflictError:
                    get_transaction().abort()
                    unsubscribed = 0
            finally:
                releaseLock(self.absolute_url())
            if not unsubscribed:
                print "Could not unsubscribe"
                time.sleep(0.5)

    def readMessagesFromTopic(self, topic, applicationId):
        '''Read published messages from a topic waiting for the application.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        '''
        try:
            getLock(self.absolute_url())
            try:
                topicOb = self.getTopic(topic)
                messages = topicOb.readMessages(applicationId)
            except ConflictResolvingTools.ConflictError:
                get_transaction().abort()
                messages = []
        finally:
            releaseLock(self.absolute_url())
        return messages

    def ackMessagesFromTopic(self, topic, applicationId, messageIds):
        '''Acknowledge that application have received messages.
        topic : case sensitive name of the topic
        applicationId : case sensitive name of the application
        Removes the messages if no subscribers are left to read.
        This MUST be called by the application, or the messages will be re-delivered.'''
        try:
            acked = 0
            while not acked:
                try:
                    getLock(self.absolute_url())
                    try:
                        topicOb = getattr(self, topic)
                        topicOb.ackMessages(applicationId, messageIds)
                        acked = 1
                    except ConflictResolvingTools.ConflictError:
                        get_transaction().abort()
                        acked = 0
                finally:
                    releaseLock(self.absolute_url())
                if not acked:
                    print "Could not ack messages."
                    time.sleep(0.5)
        except:
            tb = Helpers.GetTraceback()
            print "!"*120
            print "ackMessagesFromTopic Error"
            print "topic = %s" %topic
            print "applicationId = %s" %applicationId
            print "messageIds = %s" %messageIds
            print tb
            raise

    # Admin interface Methods
    # This method is generally only used by the administrative interface
    def setNodeId(self, nodeId, REQUEST=None):
        '''Sets this node's system-unique id.
        nodeId : system-unique id for this node
        REQUEST : Zope request object (used for redirect)
        '''
        try:
            self.nodeId = nodeId
            self._p_changed = 1
            if REQUEST:
                self.redirect(REQUEST)
        except:
            tb = Helpers.GetTraceback()
            print tb
            raise

    # This method is generally only used by the administrative interface
    def addRemoteNode(self, remoteNodeId, nodeAddress, REQUEST=None):
        '''Adds a reference to another node in the system.
        remoteNodeId : system-unique id for the remote node
        nodeAddress : XML-RPC address of the remote node (eg. http://www.foo.com/ZQueueNode0
        REQUEST : Zope request object (used for redirect)
        '''
        try:
            self.remoteNodes[remoteNodeId] = nodeAddress
            self._p_changed = 1
            if REQUEST:
                self.redirect(REQUEST)
        except:
            tb = Helpers.GetTraceback()
            print tb
            raise
    # This method is generally only used by the administrative interface
    def removeRemoteNodes(self, nodeIds, REQUEST=None):
        '''Removes references to other nodes in the system.
        nodeIds : a python list of strings containing the node Ids
        REQUEST : Zope request object (used for redirect)
        '''
        try:
            for nodeId in nodeIds:
                del self.remoteNodes[nodeId]
            self._p_changed = 1
            if REQUEST:
                self.redirect(REQUEST)
        except:
            tb = Helpers.GetTraceback()
            print tb
            raise

    tableTemplate = '''<table>
    %s
    </table>'''

    #Zope Admin interface for this object, accessed through the view tab.
    def index_html(self):
        '''Renders some info'''
        nodeId = self.getNodeId()
        remoteNodesTable = []
        for key in self.getRemoteNodes().keys():
            address = self.getRemoteNodes()[key]
            remoteNodesTable.append('''<tr>
                <td><input type="checkbox" name="nodeIds:list" value="%(key)s"></td>
                <td>%(key)s</td><td>%(address)s</td>
                </tr>''' %locals())
        remoteNodesTable = self.tableTemplate %'\n'.join(remoteNodesTable)

        currentTopics = []
        for topic in self.objectValues('ZQueueTopic'):
            currentTopics.append('''
            <a href="%s/index_html">%s</a><br>''' %(topic.absolute_url(),topic.id))
        currentTopics = self.tableTemplate %'\n'.join(currentTopics)

        currentPackets = []
        for packet in self.objectValues('ZQueuePacket'):
            currentPackets.append('''
                <a href="%s/index_html">%s</a> for topic %s<br>
                ''' %(packet.absolute_url(), packet.id, packet.getTopicId()))

        currentPackets = self.tableTemplate %'\n'.join(currentPackets)

        if self.getDeliveryHandlerIsStarted():
            deliveryHandlerStatus = 'running'
        else:
            deliveryHandlerStatus = 'stopped'

        outputOb = HTMLFile('dtml/node',globals())
        return outputOb(client=self, mapping=locals())

    # Methods called by other nodes
    def needPacket(self, packetId):
        '''Returns true if we have not yet received the packet with packetId
        packetId : unique packet id
        Returns 1 if it needs it.
        Returns 0 if it does not need it.
        Returns -1 if there was an error.
        '''
        try:
            gotValue = 0
            while not gotValue:
                try:
                    needPacket = not self.receivedPacket(packetId)
                    gotValue = 1
                except ConflictResolvingTools.ConflictError:
                    get_transaction().abort()
                    gotValue = 0
                if not gotValue:
                    print "Could not get need packet."
                    time.sleep(0.5)
            return needPacket
        except:
            get_transaction().abort()
            tb = Helpers.GetTraceback()
            print "!"*120
            print "needPacket Error"
            print "packetId = %s" %packetId
            print tb
            return -1

    def receivePacket(self, id, topic, originatorId, recipients, messages):
        '''Receive a packet of data from another node.
        id : packet Id
        topic : topic to which packet belongs
        originatorId : node id of sending node
        recipients : nodes who must receive this packet
        messages : python list of (messageId, messageBody) tuples.
        Returns 1 if the packet was received successfully
        Returns -1 if there was an error receiving the packet.'''
        try:
            createdPacket = 0
            while not createdPacket:
                try:
                    #remove ourselves from the recipients
                    del recipients[self.nodeId]
                    packet = Packet.manage_addZQueuePacket(self, id, topic, originatorId, recipients, messages)
                    self.addReceivedPacket(id)
                    #publish the messages in the local topic if this is not a QueueAdmin message
                    if topic == 'ZQueueAdminMessage':
                        self.manageAdminMessages(packet.getMessages())
                        packet.setMessagesAreDeliveredLocally()
                    createdPacket = 1
                except ConflictResolvingTools.ConflictError:
                    get_transaction().abort()
                    createdPacket = 0
                if not createdPacket:
                    print "Could not create packet."
                    time.sleep(0.5)
            return 1
        except:
            get_transaction().abort()
            tb = Helpers.GetTraceback()
            print "!"*120
            print "needPacket Error"
            print "id = %s" %id
            print "topic = %s" %topic
            print "originatorId = %s" %originatorId
            print "recipients = %s" %recipients
            print "messages = %s" %messages
            print tb
            return -1

    def packetDelivered(self, packetId, recipientId):
        '''Callback method called by receiving node to notify this node of delivery.
        packetId : packet that was delivered
        recipientId : recipient that received the packet
        Returns 1 if the callback was successfull
        Reutrns -1 if there was an error.'''
        try:
            deliveryNoted = 0
            while not deliveryNoted:
                try:
                    if hasattr(self, packetId):
                        packet = getattr(self, packetId)
                        packet.removeRecipient(recipientId)
                    deliveryNoted = 1
                except ConflictResolvingTools.ConflictError:
                    get_transaction().abort()
                    deliveryNoted = 0
                if not deliveryNoted:
                    print "Could not note delivery."
                    time.sleep(0.5)
            return 1
        except:
            get_transaction().abort()
            tb = Helpers.GetTraceback()
            print "!"*120
            print "packetDelivered Error"
            print "packetId = %s" %packetId
            print "recipientId = %s" %recipientId
            print tb
            return -1

    #internal methods
    def _p_resolveConflict(self, oldState, savedState, newState):
        '''Conflict resolver.'''
        try:
            #set up a handler with a default handler that raises an error if the values are different
            handler = ConflictResolvingTools.KeyConflictHandler(ConflictResolvingTools.errorIfDifferent)
            handler.addHandleableKey('_objects', ConflictResolvingTools.tupleMerge)
            handler.addHandleableKey('messageCounter', ConflictResolvingTools.returnLargest)
            handler.addHandleableKey('packetIdsReceived', ConflictResolvingTools.dictMerge)
            handler.addHandleableKey('remoteNodes', ConflictResolvingTools.dictMerge)
            handler.addHandleableKey('deletedRootPackets', ConflictResolvingTools.listMerge)
            retState = ConflictResolvingTools.mergeDiffs(oldState, savedState, newState, handler)
            #make sure that there are keys for every id in _objects.  Remove those that are not relevant
            ConflictResolvingTools.synchroniseObjects(retState)
            return retState
        except:
            tb = Helpers.GetTraceback()
            print "!" * 120
            print "Error resolving conflict for Node"
            print tb
            return None
           
            
##            print "NODE RESOLVER"
##
##            if hasattr(sys,'breakpoint'):
##                sys.breakpoint()
##
##            assert savedState['id'] == newState['id']
##            assert savedState['nodeId'] == newState['nodeId']
##            assert savedState['currentPacketIds'] == newState['currentPacketIds']
##            #use the largest packet counter counter
##            if savedState['messageCounter'] < newState['messageCounter']:
##                savedState['messageCounter'] = newState['messageCounter']
##
##            #apply the object diff
##            Utils.applyObjectStateDiffToState(savedState, Utils.diffObjectStates(oldState, newState))
##
##            #apply the dict diff
##            Utils.applyDictDiff(oldState['packetIdsReceived'], savedState['packetIdsReceived'], newState['packetIdsReceived'])
##            Utils.applyDictDiff(oldState['remoteNodes'], savedState['remoteNodes'], newState['remoteNodes'])
##
##            #apply the list combinations
##            Utils.combineListsUniquely(savedState['deletedRootPackets'],n['deletedRootPackets'])
##
##
##            #Do a final assertion to make sure there are no uncovered keys.
##            if __debug__:
##                assertList = ['_objects','id','pacektIdsReceived','remoteNodes','nodeId',
##                    'currentPacketIds','currentPacketNumber','deletedRootPackets','__ac_local_roles__',
##                    'deliveryHandlerStarted']
##                for d in savedState['_objects']:
##                    assertList.append(d['id'])
##
##                for key in savedState.keys():
##                    assert key in assertList, "This handler is not designed to handle key %s" %key
##            return savedState
##        except:
##            if hasattr(sys,'breakpoint'):
##                sys.breakpoint()
##            tb = Helpers.GetTraceback()
##            print tb
##            raise


    def addReceivedPacket(self, packetId):
        '''Adds the packet to the list of packets received by this node.
        packetId : id of the packet
        '''
        self.packetIdsReceived[packetId] = None
        self._p_changed = 1

    def receivedPacket(self, packetId):
        '''Returns true if we have already received this packet.
        packetId : id of the packet'''
        return self.packetIdsReceived.has_key(packetId)

    def manageAdminMessages(self, adminMessages):
        '''Handle admin messages.
        adminMessage : list of (messageId, adminMessage) tuples containing pickled admin objects.
        This method unpickles the admin objects and calls the execute method of each object.'''
        for messageId, message in adminMessages:
            sio = StringIO.StringIO(message)
            sio.seek(0)
            ob = pickle.load(sio)
            #ob must be a sub class of AdminObject
            ob.execute(self)

    def getRemoteNodes(self):
        '''Return the list remote nodes registered with this packet
        Returns a python dict of nodeId:nodeAddress values'''
        return self.remoteNodes

    def getNodeId(self):
        '''Return this node's unique nodeId'''
        return self.nodeId

    def addDeletedRootPacket(self, packetId):
        '''Add the id of a deleted root packet to the list of deleted root packets.
        packetId : id of the packet
        Once this list gets large - create a packet, notifying everybody that these
        packets are delivered.'''
        self.deletedRootPackets.append(packetId)
        if len(self.deletedRootPackets) > 10:
            self.createPacketsDeliveredNotificationPacket()
        self._p_changed = 1

    def createPacketsDeliveredNotificationPacket(self):
        '''Create a packet, notifying all the other nodes of a list of packets already deleted.'''
        messages = []
        for packetId in self.deletedRootPackets:
            ob = PacketDeletedObject(packetId)
            sio = StringIO.StringIO()
            pickle.dump(ob, sio)
            sio.seek(0)
            message = sio.read()
            messages.append(('PacketDeletedObject(%s)' %packetId, message))

        #create an admin packet
        #make a copy of our remote nodes so that the called method does not modify our dict
        remoteNodes = {}
        for key in self.getRemoteNodes().keys():
            remoteNodes[key] = 0
        packetId = "n%sZQueueAdminMessagep%s" %(self.getNodeId(), self.currentPacketNumber)
        self.currentPacketNumber += 1
        Packet.manage_addZQueuePacket(self, packetId, "ZQueueAdminMessage", None, remoteNodes, messages)
        self.deletedRootPackets = []
        self._p_changed = 1


    def createNewMessagePacket(self, topicId):
        '''Create a new message packet for a topic.
        topicId : the topic that this packet will belong to
        Packets only contain messages from one topic'''
        packetId = "n%st%sp%s" %(self.getNodeId(), topicId, self.currentPacketNumber)
        self.currentPacketNumber += 1
        #make a copy of our remote nodes so that the called method does not modify our dict
        remoteNodes = {}
        for key in self.getRemoteNodes().keys():
            remoteNodes[key] = 0

        packet = Packet.manage_addZQueuePacket(self,
                packetId, topicId, None, remoteNodes, [], 1) #we are the originators, so no originator id
        #this packet is created here, so set the delivered flag
        packet.setMessagesAreDeliveredLocally()
        #get_transaction().commit(1)
        return packetId

    def getTopic(self, topic):
        '''Return ZQueueTopic object in this node with topic name.
        topic : name of the topic
        If it doesn't exist, create it.'''
        if not hasattr(self, topic):
            Topic.manage_addZQueueTopic(self,topic)
        return getattr(self, topic)

    def getDeliveryHandler(self):
        '''Returns this node's delivery handler object.
        If none exists, create one.'''
        if not hasattr(self, 'Poller'):
            manage_addZQueuePoller(self)

        return getattr(self, 'Poller')

    def startDeliveryHandler(self, REQUEST=None):
        '''Create and Start up the delivery handler.'''

        #XXX starting must be automated in the future.

        self.getDeliveryHandler().start()
        if REQUEST:
            self.redirect(REQUEST)

    def stopDeliveryHandler(self, REQUEST=None):
        '''Stop the delivery handler and destroy it.'''
        self.getDeliveryHandler().stop()
        if REQUEST:
            self.redirect(REQUEST)

    def getDeliveryHandlerIsStarted(self):
        '''Return the running status of the delivery handler.'''
        return self.getDeliveryHandler().isRunning()

    def redirect(self, REQUEST):
        '''Redirect back to the referer.'''
        return REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


def manage_addZQueuePoller(self):
    '''Add a ZQueuePoller object to the self container.
    The object will be created with the id Poller
    This object will be created by a ZQueueNode object'''
    #XXX this object must only be created by a ZQueueNode object, we must still create security stuff to stop
    #XXX this being added from the ZMI
    self._setObject("Poller", ZQueuePoller("Poller"))

class ZQueuePoller(SimpleItem.SimpleItem):
    '''ZQueue node polling class.
    This is a worker class that does message packet creation and delivery.
    '''
    meta_type = "ZQueuePoller"

    def __init__(self, id):
        '''Initialiser.'''
        self.id = id
        
    def _p_resolveConflict(self, oldState, savedState, newState):
        '''Confilct resolver.'''
        print "!" * 120
        print " NO CONFLICT RESOLVER FOR        ZQUEUPOLLER"
        print "!" * 120
        return None
        

    def isRunning(self):
        '''Returns the running state'''
        if self.getStopEvent() is None:
            return 0
        return not self.getStopEvent().isSet()

    def start(self):
        '''Start polling.'''
        self.ownUrl = self.absolute_url()
        if not self.isRunning():
            if not hasattr(Globals, "ZQueuePoller%sEvent" %self.aq_parent.getNodeId()):
                setattr(Globals, "ZQueuePoller%sEvent" %self.aq_parent.getNodeId(), threading.Event())
            else:
                self.getStopEvent().clear()
            thread.start_new_thread(self.run,())

    def stop(self):
        '''Stop polling.'''
        if self.getStopEvent() is not None:
            self.getStopEvent().set()

    def getStopEvent(self):
        '''Get the stop event.'''
        if not hasattr(Globals, "ZQueuePoller%sEvent" %self.aq_parent.getNodeId()):
            return None
        return getattr(Globals, "ZQueuePoller%sEvent" %self.aq_parent.getNodeId())

    #XXX this method runs in a thread and calls it's own pollFolder method through XML-RPC
    #XXX as I had massive problems when "talking" directly to the ZODB
    #XXX I don't know if there is a way to "register" a thread with the ZODB so that it can
    #XXX use the objects in it directly.
    def run(self):
        '''Polling method.'''
        stopEvent = self.getStopEvent()
        while not stopEvent.isSet():
            currentPacketNumber = -1
            try:
                time.sleep(2)
                s = xmlrpclib.Server(self.ownUrl)
                currentPacketNumber = s.pollFolder()
            except:
                tb = Helpers.GetTraceback()
                print "+" * 80
                print "Poller RUN ERROR ignored."
                print "+" * 80
                print tb

    def pollFolder(self):
        '''Do a poll cycle on the parent folder.
        This method is called by the run method through XML-RPC.'''
        try:
            #make sure any transactions are comitted
            hasLock = 0
            retryCounter = 0
            while not hasLock and retryCounter < 3:
                hasLock = getLock(self.aq_parent.absolute_url(),0) #dont' wait for the lock
                if not hasLock:
##                    time.sleep(0.1)
                    retryCounter += 1
            if hasLock:
                print "=" * 80
                print "POLL CYCLE FOR %s" %self.aq_parent.id
                print "=" * 80
                delObjects = []
                #For each topic, see which messages are not packaged and package them
                for topic in self.aq_parent.objectValues('ZQueueTopic'):
                    currentPacket = None
                    for message in topic.objectValues('ZQueueMessage'):
                        if not message.isPackaged():
                            if not currentPacket:
                                currentPacketId = self.aq_parent.createNewMessagePacket(topic.id)
                                currentPacket = getattr(self.aq_parent, currentPacketId)
                            currentPacket.addMessage(message.id, message.getMessage())
                            message.setIsPackaged()
                            if message.getNumberOfSubscribersReceived() == topic.numberOfSubscribers:
                                topic.manage_delObjects([message.id])
    
                #For each packet, deliver locally and send to other nodes as needed
                for packet in self.aq_parent.objectValues('ZQueuePacket'):
                        #new packets whose contents have not been delivered locally
                    if not packet.getMessagesAreDeliveredLocally():
                        if packet.getTopicId() != 'ZQueueAdminMessage':
                            self.deliverPacketContentToLocalTopic(packet)
                        else:
                            #admin packets are handled differently and as such will always
                            #be marked as already delivered locally.
                            packet.setMessagesAreDeliveredLocally()
                    #other packets
                    else:
                        #try to deliver packets which have not yet been delivered
                        if packet.getRecipients():
                            self.deliverPacket(packet)
                        #delete packets which have been delivered
                        else:
                            self.deletePacket(packet)
                get_transaction().commit()
                releaseLock(self.aq_parent.absolute_url())
    
##                try:
##                    #########################################################################
##                    #Code to automatically deliver messages back to client.
##                    for topic in self.aq_parent.objectValues('ZQueueTopic'):
##                    #print "delivering mesgs"
##                        topic.deliverMessages()
##    
##                    ########################################################################
##                except:
##                    print "Automatic delivery code error."
##                    tb = Helpers.GetTraceback()
##                    print tb
##                    get_transaction().abort()
            else: # could not acquire the lock
                print "Could not acquire the lock, aborting poll."
        except:
            #Release the global lock, print the error and abort the transaction.
            releaseLock(self.aq_parent.absolute_url())
            tb = Helpers.GetTraceback()
            print "Poller %s Error." %(self.aq_parent.id)
            print tb
            print "-"*70
            print "Aborting transaction."
            get_transaction().abort()

    def deliverPacketContentToLocalTopic(self, packet):
        '''Deliver packet's content to local content.
        packet : the packet object to be published in local topic
        Note: only non-admin packets must be sent to this method!'''

        topic = self.aq_parent.getTopic(packet.getTopicId())
        try:
            for messageId, message in packet.getMessages():
                #add the message to the local topic
                #flag it as coming from a packet to avoid recursion
                topic.addMessage(messageId, message, 1)
            packet.setMessagesAreDeliveredLocally()
        except:
            get_transaction().abort()

    def deletePacket(self, packet):
        '''Delete the packet.
        packet : the packet object to be deleted.'''
        if packet.getOriginatorId() is None:
            #we are the root originator of this packet
            #add to the list of root packets deleted
            self.aq_parent.addDeletedRootPacket(packet.id)
        self.aq_parent.manage_delObjects(packet.id)

    def deliverPacket(self, packet):
        '''Look at the recipients and try to deliver it to them.
        packet : the packet object that must be delivered to other nodes'''
        for recipient in packet.getRecipients().keys():
            #get the recipient's address
            try:
                try:
                    address = self.aq_parent.getRemoteNodes()[recipient]
                except KeyError:
                    address = ''
                if address:
                    s = xmlrpclib.Server(address)
                    needPacket = s.needPacket(packet.id)
                    if needPacket == -1:
                        pass #
                    elif needPacket:
                        received = s.receivePacket(packet.id,
                            packet.getTopicId(),
                            self.aq_parent.getNodeId(),
                            packet.getRecipients(),
                            packet.getMessages())
                        print "Packet delivered."
                        if received != -1:
                            packet.removeRecipient(recipient)
                    else:
                        packet.removeRecipient(recipient)
                else:
                    print "No address."
            except:
                tb = Helpers.GetTraceback()
                print "Packet could not be delivered:\n%s" %tb
                #XXX figure out what all the connection type errors are that we must catch.
                pass
##            get_transaction().commit()

class AdminObject:
    '''Base class for admin objects.'''
    def execute(self, parent):
        '''Each admin object must override this method.'''
        raise NotImplementedError('Not Implemented')
        pass

#A node may not have a connection to another node.  This will cause packets that must go to that
#node to wait indefinately to be delivered.  If the packet was delivered through another route, the
#originator of that packet will eventually notify all nodes of that packet's delivery
#this admin object is the "notification" that is sent.
class PacketDeletedObject(AdminObject):
    '''This object will make sure a (usually stale) packet is marked as delivered
    so that the Poller can delete it.'''
    def __init__(self, packetId):
        self.packetId = packetId

    def execute(self, parent):
        '''See if the node contains a packet with the packet id.
        Marks it for deletion if so.'''
        if hasattr(parent, self.packetId):
            packet = getattr(parent, self.packetId)
            packet.setAllDelivered()
