from OFS import Folder
import Message
import xmlrpclib, sys
from Helpers import Helpers
from Globals import HTMLFile
import ConflictResolvingTools

def manage_addZQueueTopic(self, id):
    '''Add a ZQueueTopic to the self folder.'''
    self._setObject(id, ZQueueTopic(id))


class ZQueueTopic(Folder.Folder):
    '''ZQueue Topic Object'''
    meta_type = "ZQueueTopic"

    def __init__(self, id):
        '''Initialiser'''
        self.id = id
        self.subscribers = {}
        self.numberOfSubscribers = 0
        self.messageCounter = 0

    def publishMessage(self, message):
        '''Publishes a message in the topic.
        message must be a valid python like object'''
        #Create a unique message Id
        messageId = "n%st%sm%s" %(self.aq_parent.getNodeId(), self.id, self.messageCounter)
        self.addMessage(messageId, message)

    def calcNumberOfSubscribers(self):
        '''Calculates the number of subscribers and stores the value.'''
        self.numberOfSubscribers = len(self.subscribers.keys())
        self._p_changed = 1

    def addMessage(self, messageId, message, fromPacket=0):
        '''Add a message to the topic.'''
        self.messageCounter += 1
        self._p_changed = 1
        if self.numberOfSubscribers or not fromPacket:
            Message.manage_addZQueueMessage(self, messageId, message, fromPacket)
        #add the message to the current publishing packet
##        if not fromPacket:
##            packet = self.aq_parent.addMessageToCurrentPacket(self.id, messageId, message)

    def subscribe(self, applicationId, type=0):
        '''Subscribes to a topic.'''
        #print "Topic: subscribe. appID:",applicationId," type: ",type
        self.subscribers[applicationId] = type
        self.calcNumberOfSubscribers()
        self._p_changed = 1
        #Not sure about this line - added with conflict resolving
        #packet = self.aq_parent.addMessageToCurrentPacket(self.id, messageId, message)

    def unsubscribe(self, applicationId):
        '''Unsubscribes from topic.'''
        del self.subscribers[applicationId]
        self.calcNumberOfSubscribers()
        self._p_changed = 1

    def readMessages(self, applicationId):
        '''Read any messages waiting in topic for subscriber.
        Subscriber must ack messages to indicate that it has received them.'''
        messages = []
        for message in self.objectValues('ZQueueMessage'):
            if not message.subscriberReceived(applicationId):
                messages.append((message.id, message.getMessage()))
        return messages

    def ackMessages(self, applicationId, messageIds):
        '''Acknowledge successfull delivery of messages.
        This will:
            1) flag each message as being read by the subscriber.
            2) remove messages that have no more subscribers waiting for them.
        '''
        delObjects = []
        #print "topic:ack called"
        for messageId in messageIds:
            #print "got mesageID",messageId
            message = getattr(self, messageId)
            message.addSubsriberReceived(applicationId)
            #print "self.numbverOfSubscribers:",self.numberOfSubscribers,"\n\tmessage.subreceived: ",message.getNumberOfSubscribersReceived()
            if message.getNumberOfSubscribersReceived() == self.numberOfSubscribers and message.isPackaged():
                #print "deleting message:",messageId
                delObjects.append(messageId)
                #XXX a list must be kept of delivered messages (as all nodes will ask to deliver this message)

        if delObjects:
            self.manage_delObjects(delObjects)

    #Code to deliver messages automatically
    def deliverMessages(self):
        #print "delivering"
        for subscriber in self.subscribers.keys():
            #print "\tGot subscriber"
            if self.subscribers[subscriber] == 1:
                #print "\t\tNon-polling"
                #If not polling send msg
                messages = []
                for message in self.objectValues('ZQueueMessage'):
                    #print "found msg"
                    if not message.subscriberReceived(subscriber):
                        messages.append((message.id, message.getMessage()))
                if messages:
                    try:
                        print "Topic:Non-polling trying callback"
                        xmlrpclib.Server(subscriber).callback(self.id, messages)
                    except:
                        print "xmlrpc connction error trying to callback"
                        pass


    def index_html(self):
        '''Admin interface for topic.'''
        topicId = self.id
        subscriberList = []
        for subscriber in self.subscribers.keys():
            subscriberList.append('''
            %s<br>''' %subscriber)

        subscriberList = "\n".join(subscriberList)

        messageList = []
        for message in self.objectValues('ZQueueMessage'):
            messageList.append('''
            <a href="%s/index_html">%s</a><br>''' %(message.absolute_url(),message.id))

        messageList = "\n".join(messageList)

        return HTMLFile('dtml/topic', globals())(client=self, mapping=locals())


    def _p_resolveConflict(self, oldState, savedState, newState):
        '''Conflict resolver.'''
        try:
            #set up a handler with a default handler that raises an error if the values are different
            handler = ConflictResolvingTools.KeyConflictHandler(ConflictResolvingTools.errorIfDifferent)
            handler.addHandleableKey('_objects', ConflictResolvingTools.tupleMerge)
            handler.addHandleableKey('messageCounter', ConflictResolvingTools.returnLargest)
            handler.addHandleableKey('subscribers', ConflictResolvingTools.dictMerge)
            handler.addHandleableKey('numberOfSubscribers', ConflictResolvingTools.returnLargest)
            retState = ConflictResolvingTools.mergeDiffs(oldState, savedState, newState, handler)
            retState['numberOfSubscribers'] = len(retState['subscribers'].keys())
            #make sure that there are keys for every id in _objects.  Remove those that are not relevant
            ConflictResolvingTools.synchroniseObjects(retState)
            return retState
        except:
            tb = Helpers.GetTraceback()
            print "!" * 120
            print "Error resolving conflict for Topic"
            print tb
            return None
